//STARTHEADER
// $Id: HEPTopTagger.cc v1.3 2014-12-03 by M.Takeuchi $
//----------------------------------------------------------------------
//ENDHEADER

#include <fastjet/tools/HEPTopTagger.hh>
#include <fastjet/tools/Filter.hh>
#include <fastjet/tools/Pruner.hh>
#include <fastjet/Error.hh>
#include <fastjet/JetDefinition.hh>
#include <fastjet/ClusterSequence.hh>
#include <sstream>

FASTJET_BEGIN_NAMESPACE

using namespace std;

//----------------------------------------------------------------------
// HEPTopTagger class implementation
//----------------------------------------------------------------------

LimitedWarning HEPTopTagger::_warnings_nonca;

//------------------------------------------------------------------------
// description of the tagger
string HEPTopTagger::description() const{ 
  ostringstream oss;
  oss << "HEPTopTagger (mt=" << _mt << ", mW=" << _mW << ")";
  if(_return_only_tagged){oss << " return only tagged jet"<< endl;}
  else{oss << " return also non-tagged jet with 3 subjets found"<< endl;}
  if(_check_massratio_first){oss << "--- only consider triplet passing massratio cut"<< endl;}
  oss << " mass_drop_threshold= " << _mass_drop_threshold
      << ", max_subjet_mass= " << _max_subjet_mass << endl;
  oss << " (mtmin, mtmax)=( " << _mtmin
      << ", " << _mtmax << ")" << endl;
  oss << " (rmin,rmax)=( " << _rmin
      << ", " << _rmax << ")" << endl;
  oss << " (m23min, m13min, m13max)=(" 
      << _m23cut << "," <<  _m13cutmin << "," << _m13cutmax << ")";
  oss << description_of_selectors();
  return oss.str();
}

string HEPTopTagger::description_filter() const{ 
  ostringstream oss;
  oss << "HEPTopTagger Filtering:" << endl;
  oss << " nfilter=" << _nfilt
      << ", jetalgorithm_filter=" << _jet_algorithm_filter << endl;
  oss<< " jetalgorithm_recluster=" << _jet_algorithm_recluster;
  return oss.str();
}

string HEPTopTagger::description_prune() const{ 
  ostringstream oss;
  oss << "HEPTopTagger Pruning:" << endl;
  oss << " zcut=" << _zcut 
      << " rcutfac=" << _rcutfac
      << ", jetalgorithm_prune=" << _jet_algorithm_prune;
  return oss.str();
}


//------------------------------------------------------------------------
// returns the tagged PseudoJet if 3 subjets found, 0 otherwise
// tag is need to use.
PseudoJet HEPTopTagger::result(const PseudoJet & jet) const{
  if (!jet.has_valid_cluster_sequence()){
    throw Error("HEPTopTagger can only be applied on jets having an associated (and valid) ClusterSequence");
  }
  // warn if the jet has not been clustered with a Cambridge/Aachen algorithm
  if (jet.validated_cs()->jet_def().jet_algorithm() != cambridge_algorithm)
    _warnings_nonca.warn("HEPTopTagger should only be applied on jets from a Cambridge/Aachen clustering; use it with other algorithms at your own risk.");
  
  vector<PseudoJet> allsubjets;
  vector<PseudoJet> rejected_by_massdrop;
  FindHardSubst(jet,allsubjets,rejected_by_massdrop);
  //cout << allsubjets.size() << " subjets found"<<endl;
  if(allsubjets.size()<3) return PseudoJet();

  double _deltatop = numeric_limits<double>::max();

  PseudoJet topcand_unfiltered;
  PseudoJet topcand_filtered;
  vector<unsigned> selected;
  for(unsigned rr=0; rr<allsubjets.size(); rr++){
    for(unsigned ll=rr+1; ll<allsubjets.size(); ll++){
      for(unsigned kk=ll+1; kk<allsubjets.size(); kk++){
	PseudoJet topcand_tmp= join(allsubjets[rr],allsubjets[ll],
				    allsubjets[kk]);
	PseudoJet topcand_filtered_tmp;
	double deltatop_tmp=Distance(topcand_tmp,topcand_filtered_tmp);
	
	if(deltatop_tmp<_deltatop){
	  topcand_unfiltered = topcand_tmp;
	  topcand_filtered = topcand_filtered_tmp;
	  _deltatop = deltatop_tmp;
	  selected.clear();
	  selected.push_back(rr);
	  selected.push_back(ll);
	  selected.push_back(kk);
	}
      }// end kk
    }// end ll
  }// end rr
  if(_deltatop ==  numeric_limits<double>::max()) return PseudoJet();

  //cout << "after triplet choice"<<endl;  
  vector<PseudoJet> rejected_by_filter=topcand_filtered.structure_of<Filter>().rejected(); 
  vector<PseudoJet> not_selected;
  for(unsigned rr=0; rr<allsubjets.size(); rr++){
    if(rr!=selected.at(0) && rr!=selected.at(1) && rr!=selected.at(2))
      not_selected.push_back(allsubjets.at(rr));
  }

  //reclustering
  JetDefinition jetdef_recluster(_jet_algorithm_recluster, 3.14);
  ClusterSequence* cs_rec=new ClusterSequence(topcand_filtered.constituents(),jetdef_recluster);
  vector<PseudoJet> top_rec= sorted_by_topkinematics(cs_rec->exclusive_jets(3));
  cs_rec->delete_self_when_unused(); // to keep ClusterSequence

  // create the result and its structure
  const JetDefinition::Recombiner* rec
    = jet.associated_cluster_sequence()->jet_def().recombiner();
  
  PseudoJet W = join(top_rec[1], top_rec[2],*rec);
  PseudoJet non_W = join(top_rec[0],*rec);
  PseudoJet result_local = join<HEPTopTaggerStructure>(W, non_W,*rec);

  // Pruning
  Pruner pruner(_jet_algorithm_prune,_zcut,_rcutfac);
  PseudoJet topcand_pruned = pruner(topcand_unfiltered);
  
  HEPTopTaggerStructure *s = (HEPTopTaggerStructure*) result_local.structure_non_const_ptr();
  s->_cos_theta_w = _cos_theta_W(result_local);
  s->_is_masscut_passed = check_mass_criteria(top_rec);
  s->_is_massrange_passed = _mtmin<result_local.m() && result_local.m()< _mtmax;
  s->_unfiltered= topcand_unfiltered;
  s->_pruned= topcand_pruned;
  s->_rejected_by_massdrop= rejected_by_massdrop;
  s->_rejected_by_filter= rejected_by_filter;
  s->_not_selected= not_selected;
  s->_delta_top= _deltatop;
  
  if(_return_only_tagged && !s->is_tagged() ) result_local=PseudoJet();
  return result_local; 
}




// HEPTopTagger decomposition procedure
void HEPTopTagger::FindHardSubst(const PseudoJet & this_jet,vector <PseudoJet>  & allsubjets,vector <PseudoJet>  & rejected_by_massdrop) const
{
  PseudoJet p1, p2;
  if (this_jet.m() < _max_subjet_mass || !this_jet.validated_cs()->has_parents(this_jet,p1,p2))
    {
      allsubjets.push_back(this_jet);
    }
  else
    {
      if (p1.m()<p2.m()) std::swap(p1, p2);   
      FindHardSubst(p1,allsubjets,rejected_by_massdrop);
      if (p1.m()<_mass_drop_threshold*this_jet.m()){
	FindHardSubst(p2,allsubjets,rejected_by_massdrop);
      }
      else{
	rejected_by_massdrop.push_back(p2);
      }
    }
}

//------------------------------
double HEPTopTagger::Distance(PseudoJet & topcandtmp, PseudoJet & topcand_filtered) const{
  vector<PseudoJet> pieces=topcandtmp.pieces();
  // define Filtering: filt_top_R and jetdefinition
  double Rfilt_top
    = min(0.3,0.5*sqrt(min(pieces[0].squared_distance(pieces[1]),
			   min(pieces[2].squared_distance(pieces[1]),
			       pieces[0].squared_distance(pieces[2])))));
  JetDefinition jetdef_filter(_jet_algorithm_filter, Rfilt_top);
  Selector selector_filter=SelectorNHardest(_nfilt);
  Filter filter(jetdef_filter,selector_filter);
  PseudoJet cand_filtered = filter(topcandtmp);
  topcand_filtered=cand_filtered;
  double deltatop = abs(cand_filtered.m() - _mt);

  if(_check_massratio_first){
    JetDefinition jetdef_recluster(_jet_algorithm_recluster, 3.14);
    ClusterSequence cs_rec(topcand_filtered.constituents(),jetdef_recluster);
    if(!check_mass_criteria(sorted_by_topkinematics(cs_rec.exclusive_jets(3))))
      deltatop = numeric_limits<double>::max();
  }
  return deltatop;
}



vector <PseudoJet> HEPTopTagger::sorted_by_topkinematics(const vector<PseudoJet>& top_subs_in) const{
  if(top_subs_in.size()!=3){
    throw Error("sorted_by_topkinematics is only applicable for vector<PseudoJet> with 3 entries");
    vector <PseudoJet> vec_0;
    return vec_0;
  }
  vector <PseudoJet> sorted;
  vector <PseudoJet> top_subs=sorted_by_pt(top_subs_in);
  double dm12=abs((top_subs[0]+top_subs[1]).m()-_mW);
  double dm13=abs((top_subs[0]+top_subs[2]).m()-_mW);
  double dm23=abs((top_subs[1]+top_subs[2]).m()-_mW);
  if(dm23<=dm12 && dm23<=dm13){
    sorted.push_back(top_subs[0]); //supposed to be b
    sorted.push_back(top_subs[1]); //W-jet 1
    sorted.push_back(top_subs[2]); //W-jet 2		
  }
  else if(dm13<=dm12 && dm13<dm23){
    sorted.push_back(top_subs[1]); //supposed to be b
    sorted.push_back(top_subs[0]); //W-jet 1
    sorted.push_back(top_subs[2]); //W-jet 2
  }
  else if(dm12<dm23 && dm12<dm13){
    sorted.push_back(top_subs[2]); //supposed to be b
    sorted.push_back(top_subs[0]); //W-jet 1
    sorted.push_back(top_subs[1]); //W-jet 2
  }
  return sorted;
}



bool HEPTopTagger::check_mass_criteria(const vector<PseudoJet> & top_subs) const{
  bool is_passed=false;
  double m12=(top_subs[0]+top_subs[1]).m();
  double m13=(top_subs[0]+top_subs[2]).m();
  double m23=(top_subs[1]+top_subs[2]).m();
  double m123=(top_subs[0]+top_subs[1]+top_subs[2]).m();
  if(
     (atan(m13/m12)>_m13cutmin && _m13cutmax > atan(m13/m12)
      && (m23/m123>_rmin && _rmax>m23/m123))
     ||
     (((m23/m123)*(m23/m123) < 1-_rmin*_rmin*(1+(m13/m12)*(m13/m12))) &&
      ((m23/m123)*(m23/m123) > 1-_rmax*_rmax*(1+(m13/m12)*(m13/m12))) && 
      (m23/m123 > _m23cut))
     ||
     (((m23/m123)*(m23/m123) < 1-_rmin*_rmin*(1+(m12/m13)*(m12/m13))) &&
      ((m23/m123)*(m23/m123) > 1-_rmax*_rmax*(1+(m12/m13)*(m12/m13))) && 
      (m23/m123 > _m23cut))
     ){ 
    is_passed=true;
  }
  return is_passed;
}



FASTJET_END_NAMESPACE

