#ifndef __FASTJET_HEP_TOP_TAGGER_HH__
#define __FASTJET_HEP_TOP_TAGGER_HH__

//STARTHEADER
// $Id: HEPTopTagger.hh XXXX 2012-11-22 12:25: M.Takeuchi v1.3 $
//----------------------------------------------------------------------
//ENDHEADER

#include <fastjet/tools/TopTaggerBase.hh>
#include <fastjet/CompositeJetStructure.hh>
#include <fastjet/LimitedWarning.hh>

FASTJET_BEGIN_NAMESPACE

class HEPTopTagger;
class HEPTopTaggerStructure;

//----------------------------------------------------------------------
/// @ingroup tools_taggers
/// \class HEPTopTagger
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
/// Class that helps perform boosted top tagging using the "HEPTopTagger"
/// in arXiv:1006.2833 (T. Plehn, M. Spannowsky, M. Takeuchi and D. Zerwas)
///
///The tagger proceeds as follows:
///  1. start from a jet J obtained with the Cambridge/Aachen algorithm
///  2. undo the last iteration J -> j_1,j_2 (with m_j1>m_j2) until the
///     subjet j1 satisfies m_j1 < 0.8*m_J
///   - if the criteria is not satisfied, carry on the procedure with 
///     j_1 (discarding j_2)
///   - if the criteria m_j1 < 0.8*m_J is satisfied, keep both j_1 and j_2
///     and repeat the procedure
///   - if the criteria m_j1 < 30 GeV (now the default value is 50 GeV) 
///     is satisfied, keep j_1 as a subjet
///   - if less than 3 subjets found, the tagger has failed
///  3. taking the 3 subjets which give the best top mass after filtering
///     and check if it is consistent to the top mass
///  4. recluster the filtered jets into 3 subjets
///  5. reconstruct the three invariant masses from the 3 subjets 
///  6. impose the mass ratio cuts on them
///
/// \section Input conditions
/// 
///  - the original jet must have an associated (and valid)
///    ClusterSequence
///  - the tagger is designed to work with jets formed by the
///    Cambridge/Aachen (C/A) algorithm; if a non-C/A jet is passed to
///    the tagger, a warning will be issued
///
/// \section Example
///
/// A  HEPTopTagger can be used as follows:
///
/// \code
///    double max_subjet_mass = 50.; // subjets mass threthold
///    HEPTopTagger heptoptagger(max_subjet_mass);
///    // indicate the acceptable range of top, W masses (default: no limits)
///    top_tagger.set_mass_ratio_cut(0.35,0.2,1.3);
///    // now try and tag a jet
///    PseudoJet top_candidate = heptoptagger(jet);  // jet should come from a Cambridge/Aachen clustering
///    if (top_candidate != PseudoJet()) { // top candidate found
///      double top_mass = top_candidate.m();
///      double W_mass = top_candidate.structure_of<HEPTopTagger>().W().m();
///      bool is_top_tagged=topcandidate.structure_of<HEPTopTagger>().is_tagged();
///      if(is_top_tagged) cout << "it is a tagged top" << endl; 
///    }
/// \endcode
///
/// The full set of information available from the structure_of<HEPTopTagger>() 
/// call is
///
/// - PseudoJet W()    : the W subjet of the top candidate
/// - PseudoJet non_W(): non-W subjet(s) of the top candidate (i.e. the b)
/// - double cos_theta_W(): the W helicity angle
/// - PseudoJet W1(): the harder of the two prongs of the W
/// - PseudoJet W2(): the softer of the two prongs of the W
/// - vector<PseudoJet> rejected_by_massdrop(): the subjets rejected by mass drop criteria
/// - vector<PseudoJet> rejected_by_filter(): the filtered jets rejected by filtering procedure
/// - vector<PseudoJet> not_selected(): the subjets found by mass drop criteria but not selected for top reconstruction
/// The structure of the top_candidate can also be accessed through its
/// pieces() function:
///
/// - top_candidate.pieces()[0]: W
/// - top_candidate.pieces()[1]: non_W
///
/// The W itself has two pieces (corresponding to W1, W2). 
///
/// The existence of the first two of the structural calls (W(),
/// non_W()) and the fact that the top is made of two pieces (W,
/// non_W) are features that should be common to all taggers derived
/// from TopTaggerBase.
///
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
///
class HEPTopTagger : public TopTaggerBase {
public:
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
  /// default ctor
  /// The parameters are the following:
  ///  \param max_subjet_mass      threshold of the subjet mass
  ///  \param mass_drop_threshold  threshold fraction for mass drop criteria
  ///  \param mt                   the top mass
  ///  \param mW                   the W mass
  ///  \param mtmin                the lower threshold of the top range
  ///  \param mtmax                the upper threshold of the top range
  ///
  /// The default values of all these parameters are taken from
  /// arXiv:1006.2833
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
////////// EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE /////////
  HEPTopTagger(const double max_subjet_mass=50.,
	       const double mass_drop_threshold=0.8, 
	       const double mt=172.4,const double mW=80.4, 
	       const double mtmin=150.,const double mtmax=200.)
    :_max_subjet_mass(max_subjet_mass),
     _mass_drop_threshold(mass_drop_threshold),
     _mt(mt),_mW(mW),
     _mtmin(mtmin),_mtmax(mtmax),
     _rmin(0.85*mW/mt),_rmax(1.15*mW/mt),
     _m23cut(0.35),_m13cutmin(0.2),_m13cutmax(1.3),
     _nfilt(5),_zcut(0.1),_rcutfac(0.5),
     _jet_algorithm_filter(cambridge_algorithm),
     _jet_algorithm_recluster(cambridge_algorithm),
     _jet_algorithm_prune(cambridge_algorithm),
     _check_massratio_first(false),
     _return_only_tagged(false){}
  
  /// returns a textual description of the tagger
  virtual std::string description() const;
  virtual std::string description_filter() const;
  virtual std::string description_prune() const;
  
  /// runs the tagger on the given jet and
  /// returns the tagged PseudoJet if successful, or a PseudoJet==0 otherwise
  /// (standard access is through operator()).
  ///  \param jet   the PseudoJet to tag
  virtual PseudoJet result(const PseudoJet & jet) const;

  void FindHardSubst(const PseudoJet & this_jet, std::vector < PseudoJet >& allsubjets,std::vector<PseudoJet>& removed_by_massdrop) const;
  void set_max_subjet_mass(double x) {_max_subjet_mass=x;}
  void set_mass_drop_threshold(double x) {_mass_drop_threshold=x;}
  void set_top_range(double xmin, double xmax) {_mtmin=xmin; _mtmax=xmax;}
  void set_mass_ratio_range(double rmin, double rmax) {_rmin=rmin; _rmax=rmax;}
  void set_mass_ratio_cut(double m23cut, double m13cutmin,double m13cutmax){_m23cut=m23cut; _m13cutmin=m13cutmin; _m13cutmax=m13cutmax;}
  void set_nfilt(unsigned nfilt) {_nfilt=nfilt;}
  void set_prun_zcut(double zcut) {_zcut=zcut;}
  void set_prun_rcutfac(double rcutfac) {_rcutfac=rcutfac;}
  void set_filter_jetalgorithm(JetAlgorithm jet_algorithm) {_jet_algorithm_filter=jet_algorithm;}
  void set_recluster_jetalgorithm(JetAlgorithm jet_algorithm) {_jet_algorithm_recluster=jet_algorithm;}
  void set_prune_jetalgorithm(JetAlgorithm jet_algorithm) {_jet_algorithm_prune=jet_algorithm;}
  void check_massratio_first(bool on_off) {_check_massratio_first=on_off;}
  void return_only_tagged(bool on_off) {_return_only_tagged=on_off;}
  
  // the type of the associated structure
  typedef HEPTopTaggerStructure StructureType;
  
protected:
  /// runs the HEPTopTagger decomposition procedure
  double Distance(PseudoJet & topcandtmp, PseudoJet & topfilteredtmp) const;
  //PseudoJet GetFiltered(PseudoJet & topcandtmp, double & deltatop) const;

  //private:
  double _max_subjet_mass; // stop when subjet mass < 50 GeV
  double _mass_drop_threshold;
  const double _mt, _mW;
  double _mtmin, _mtmax;
  double _rmin, _rmax;
  double _m23cut, _m13cutmin, _m13cutmax;
  size_t _nfilt;
  double _zcut;
  double _rcutfac;
  // JetAlgorithms
  JetAlgorithm _jet_algorithm_filter;
  JetAlgorithm _jet_algorithm_recluster;
  JetAlgorithm _jet_algorithm_prune;
  bool _check_massratio_first;
  bool _return_only_tagged;

  std::vector < PseudoJet > sorted_by_topkinematics(const std::vector < PseudoJet >& top_subs) const;
  bool check_mass_criteria(const std::vector < fastjet::PseudoJet > & top_subs) const;
  //PseudoJet Sum(const vector<PseudoJet>& );

  static LimitedWarning _warnings_nonca;
};


//------------------------------------------------------------------------
/// @ingroup tools_taggers
/// \class HEPTopTaggerStructure
/// the structure returned by the HEPTopTagger transformer.
///
/// See the HEPTopTagger class description for the details of what
/// is inside this structure
///
class HEPTopTaggerStructure : public CompositeJetStructure, public TopTaggerBaseStructure {
public:
  /// ctor with pieces initialisation
  HEPTopTaggerStructure(std::vector<PseudoJet> pieces_in,
                 const JetDefinition::Recombiner *recombiner = 0) :
    CompositeJetStructure(pieces_in, recombiner),
    _is_massrange_passed(false),_is_masscut_passed(false){}

  inline const bool is_tagged() const{
	  return _is_massrange_passed && _is_masscut_passed;
  }

  inline const bool is_massrange_passed() const{
	  return _is_massrange_passed;
  }

  inline const bool is_masscut_passed() const{
	  return _is_masscut_passed;
  }


  inline const double delta_top() const{
	  return _delta_top;
  }

  /// returns the W subjet
  inline const PseudoJet & W() const{ 
    return _pieces[0];
  }

  /// returns the first W subjet (the harder)
  inline PseudoJet W1() const{
    assert(W().pieces().size()>0);
    return W().pieces()[0];
  }
  
  /// returns the second W subjet
  inline PseudoJet W2() const{
    assert(W().pieces().size()>1);
    return W().pieces()[1];
  }

  /// returns the non-W subjet
  /// It will have 1 or 2 pieces depending on whether the tagger has
  /// found 3 or 4 pieces
  inline const PseudoJet & non_W() const{ 
    return _pieces[1];
  }

  /// returns the W helicity angle
  inline double cos_theta_W() const {return _cos_theta_w;}

//  returns the grooming results
  inline const PseudoJet & unfiltered() const {return _unfiltered;}
  inline const PseudoJet & pruned() const {return _pruned;}
  
  inline const vector<PseudoJet> & rejected_by_filter() const{return _rejected_by_filter;}
  inline const vector<PseudoJet> & rejected_by_massdrop() const{return _rejected_by_massdrop;}
  inline const vector<PseudoJet> & not_selected() const{return _not_selected;}
  
protected:
  double _cos_theta_w;      ///< the W helicity angle
  //PseudoJet _W;             ///< the tagged W
  //PseudoJet _non_W;         ///< the remaining pieces
  //  PseudoJet _original_jet;  ///< the original jet (before tagging)
  bool _is_massrange_passed;
  bool _is_masscut_passed;
  double _delta_top;
  unsigned _top_count;
  //unsigned _parts_size;
  PseudoJet _unfiltered;
  PseudoJet _pruned;
  vector<PseudoJet> _rejected_by_filter;
  vector<PseudoJet> _rejected_by_massdrop;
  vector<PseudoJet> _not_selected;

  //vector<PseudoJet> _top_subjets;
  //vector<PseudoJet> _top_hadrons;
  //vector<PseudoJet> _top_parts;
  //vector<vector<PseudoJet> > _candjets;

  double check_cos_theta(const PseudoJet & jet, const PseudoJet & subj1,const PseudoJet & subj2) const;
  
  double r_max_3jets(const fastjet::PseudoJet & jet1,const fastjet::PseudoJet & jet2,
		     const fastjet::PseudoJet & jet3) const;

  // allow the tagger to set these
  friend class HEPTopTagger;
};



FASTJET_END_NAMESPACE
#include "HEPTopTagger.cc"


#endif  //  __FASTJET_HEP_TOP_TAGGER_HH__

